# Description: This file contains the EventProcessor class which is used to process the event
# received from the listener.
from django.utils import timezone


class DefaultProcessor:
    def __init__(self, event, force_process=False):
        self.event = event
        self.force_process = force_process

    def process(self):
        event = self.event
        if self.event.processed and not self.force_process:
            return
        event.processed = True
        event.processed_at = timezone.now()
        event.save()



from django.conf import settings


class Settings:
    # Contains the settings for the listener app, check django.conf.settings first
    PROCESSORS = {
        "default": "listener.processor.DefaultProcessor",
    }

    def __init__(self):
        django_settings = getattr(settings, "LISTENER", {})
        for key, value in django_settings.items():
            setattr(self, key, value)



# Generated by Django 4.2.6 on 2023-11-02 10:35

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Source',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('name', models.CharField(max_length=255)),
                ('slug', models.SlugField(max_length=255, unique=True)),
                ('url', models.URLField(max_length=1000)),
                ('is_active', models.BooleanField(default=True)),
                ('unique_id_field', models.CharField(default='id', max_length=255)),
                ('username', models.CharField(blank=True, max_length=255, null=True)),
                ('password', models.CharField(blank=True, max_length=255, null=True)),
                ('token', models.TextField(blank=True, null=True)),
                ('token_header', models.CharField(default='HTTP_AUTHORIZATION', max_length=255)),
                ('http_method', models.CharField(default='POST', max_length=255)),
                ('last_checked', models.DateTimeField(blank=True, null=True)),
                ('content_type', models.CharField(default='application/json', max_length=255)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Event',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('name', models.CharField(blank=True, max_length=255, null=True)),
                ('body', models.TextField(blank=True, null=True)),
                ('headers', models.TextField(blank=True, null=True)),
                ('query_params', models.TextField(blank=True, null=True)),
                ('unique_id', models.CharField(blank=True, db_index=True, max_length=255, null=True)),
                ('processed', models.BooleanField(default=False)),
                ('processed_at', models.DateTimeField(blank=True, null=True)),
                ('source', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='events',
                                             to='listener.source')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



import base64

from django.test import TestCase

from listener.models import Source, Event


# Create your tests here.
# Path: listener/tests.py


class SourceModelTestCase(TestCase):
    def setUp(self):
        self.source = Source.objects.create(
            name="Test Source",
            slug="test-source",
            url="https://example.com/webhooks/test-source/",
            is_active=True,
            unique_id_field="id",
            username="test",
            password="test",
            token="test",
            token_header="HTTP_AUTHORIZATION",
            http_method="POST",
            content_type="application/json",
        )

    def test_source_str(self):
        self.assertEqual(str(self.source), "Test Source")

    def test_source_save(self):
        self.assertEqual(self.source.slug, "test-source")
        self.assertTrue(self.source.authenticate("test"))
        self.assertFalse(self.source.authenticate("invalid"))

    def test_source_authenticate(self):
        self.assertTrue(self.source.authenticate("test"))
        self.assertFalse(self.source.authenticate("invalid"))

    def test_source_authenticate_token(self):
        self.source.username = None
        self.source.password = None
        self.source.save()
        self.assertTrue(self.source.authenticate("test"))
        self.assertFalse(self.source.authenticate("invalid"))

    def test_source_authenticate_no_auth(self):
        self.source.username = None
        self.source.password = None
        self.source.token = None
        self.source.save()
        self.assertTrue(self.source.authenticate("test"))
        self.assertTrue(self.source.authenticate("invalid"))
    
    def test_get_processor_class(self):
        from listener.settings import Settings
        from listener.processor import DefaultProcessor
        self.assertEqual(self.source.get_processor_class(), DefaultProcessor)


class WebhookViewTestCase(TestCase):
    def setUp(self):
        self.source = Source.objects.create(
            name="Test Source",
            slug="test-source",
            url="https://example.com/webhooks/test-source/",
            is_active=True,
            unique_id_field="id",
            username="test",
            password="test",
            token="test",
            token_header="HTTP_AUTHORIZATION",
            http_method="POST",
            content_type="application/json",
        )

    def test_webhook_view(self):
        response = self.client.get("/webhooks/test-source/")
        self.assertEqual(response.status_code, 405)
        response = self.client.post("/webhooks/test-source/")
        self.assertEqual(response.status_code, 401)
        credentials = base64.b64encode(b"test:test").decode("utf-8")
        headers = {"Host": "testserver", "Connection": "close", "Content-Type": "application/json"}
        response = self.client.post(
            "/webhooks/test-source/", HTTP_AUTHORIZATION="Basic %s" % credentials, content_type="application/json",
            data={"test": "test", "id": "test"}, headers=headers
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(Event.objects.count(), 1)
        event = Event.objects.first()
        self.assertEqual(event.source, self.source)
        self.assertEqual(event.name, "Test Source")
        self.assertEqual(event.body, '{"test": "test", "id": "test"}')
        self.assertEqual(event.query_params, '{}')
        self.assertEqual(event.unique_id, "test")

    def test_webhook_view_token(self):
        self.source.username = None
        self.source.password = None
        self.source.save()
        response = self.client.post("/webhooks/test-source/", HTTP_AUTHORIZATION="Bearer dGVzdDp0ZXN0",
                                    content_type="application/json", data={"test": "test", "id": "test"})
        self.assertEqual(response.status_code, 401)
        response = self.client.post("/webhooks/test-source/", HTTP_AUTHORIZATION="test",
                                    content_type="application/json", data='{"test": "test", "id": "test"}')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(Event.objects.count(), 1)
        event = Event.objects.first()
        self.assertEqual(event.source, self.source)
        self.assertEqual(event.name, "Test Source")
        self.assertEqual(event.body, '{"test": "test", "id": "test"}')
        self.assertEqual(event.query_params, '{}')
        self.assertEqual(event.unique_id, "test")
        

class DefaultProcessorTestCase(TestCase):
    def setUp(self):
        self.source = Source.objects.create(
            name="Test Source",
            slug="test-source",
            url="https://example.com/webhooks/test-source/",
            is_active=True,
            unique_id_field="id",
            username="test",
            password="test",
            token="test",
            token_header="HTTP_AUTHORIZATION",
            http_method="POST",
            content_type="application/json",
        )
        self.event = Event.objects.create(
            source=self.source,
            name="Test Source",
            body='{"test": "test", "id": "test"}',
            headers='{"Host": "testserver", "Connection": "close", "Content-Type": "application/json"}',
            query_params='{}',
            unique_id="test",
        )

    def test_process(self):
        processor = self.source.get_processor_class()(self.event)
        self.assertEqual(processor.process(), None)
        self.event.refresh_from_db()
        self.assertEqual(self.event.processed, True)
        self.assertNotEqual(self.event.processed_at, None)



from django.contrib import admin

from listener.models import Source, Event


# Register your models here.


@admin.register(Source)
class SourceAdmin(admin.ModelAdmin):
    list_display = ('name', 'slug', 'url', 'is_active', 'last_checked', 'content_type',)
    list_filter = ('is_active', 'content_type',)
    search_fields = ('name', 'slug', 'url', 'username', 'password',)
    prepopulated_fields = {'slug': ('name',)}
    readonly_fields = ('last_checked', 'created', 'modified',)


@admin.register(Event)
class EventAdmin(admin.ModelAdmin):
    list_display = ('__str__', 'name', 'unique_id', 'created', 'processed',)
    list_filter = ('source',)
    search_fields = ('source__name', 'name', 'unique_id',)
    readonly_fields = ('created', 'modified',)
    actions = ['process_event', ]

    def process_event(modeladmin, request, queryset):
        for obj in queryset:
            source = obj.source
            processor_class = source.get_processor_class()
            processor = processor_class(obj, force_process=True)
            processor.process()

    process_event.short_description = "Process selected events"



from importlib import import_module

from django.contrib.auth.hashers import make_password, check_password
from django.db import models
from django.utils.text import slugify


# Create your models here.

class BaseModel(models.Model):
    created = models.DateTimeField(auto_now_add=True)
    modified = models.DateTimeField(auto_now=True)

    class Meta:
        abstract = True


class Source(BaseModel):
    name = models.CharField(max_length=255)
    slug = models.SlugField(max_length=255, unique=True)
    url = models.URLField(max_length=1000)
    is_active = models.BooleanField(default=True)
    unique_id_field = models.CharField(max_length=255,
                                       default='id')  # use (dot) notation for nested fields, e.g. 'data.id'

    username = models.CharField(max_length=255, blank=True, null=True)
    password = models.CharField(max_length=255, blank=True, null=True)
    token = models.TextField(blank=True, null=True)
    token_header = models.CharField(max_length=255, default='HTTP_AUTHORIZATION')

    http_method = models.CharField(max_length=255, default='POST')

    last_checked = models.DateTimeField(null=True, blank=True)
    content_type = models.CharField(max_length=255, default='application/json')

    def __str__(self):
        return self.name

    def save(self, *args, **kwargs):
        if not self.slug:
            self.slug = slugify(self.name)
        if self.password:
            # encrypt password
            self.password = make_password(self.password)
        super().save(*args, **kwargs)

    def authenticate(self, raw_password):
        if self.username and self.password:
            return check_password(raw_password, self.password)
        elif self.token:
            return raw_password == self.token
        return True

    def get_processor_class(self):
        from listener.settings import Settings
        from listener.processor import DefaultProcessor

        settings = Settings()
        processor_class = settings.PROCESSORS.get(self.slug, settings.PROCESSORS.get("default"))
        # Import the processor class
        if not processor_class:
            return DefaultProcessor
        try:
            module_name, class_name = processor_class.rsplit(".", 1)
            module = import_module(module_name)
            return getattr(module, class_name)
        except (ImportError, AttributeError):
            return DefaultProcessor


class Event(BaseModel):
    source = models.ForeignKey(Source, on_delete=models.CASCADE, related_name="events")
    name = models.CharField(max_length=255, blank=True, null=True)
    body = models.TextField(blank=True, null=True)
    headers = models.TextField(blank=True, null=True)
    query_params = models.TextField(blank=True, null=True)
    unique_id = models.CharField(max_length=255, blank=True, null=True, db_index=True)

    processed = models.BooleanField(default=False)
    processed_at = models.DateTimeField(blank=True, null=True)

    def __str__(self):
        return "{} - {}".format(self.source, self.unique_id)



import base64


def authenticate(request, source):
    if source.username and source.password:
        # Basic auth
        auth_header = request.META.get(source.token_header) or request.META.get('HTTP_AUTHORIZATION')
        if not auth_header:
            return False
        encoded_credentials = auth_header.split(' ')[1]
        decoded_credentials = base64.b64decode(encoded_credentials).decode("utf-8").split(':')
        username = decoded_credentials[0]
        password = decoded_credentials[1]
        return source.username == username and source.authenticate(password)
    elif source.token:
        # Token auth
        return source.authenticate(request.META.get(source.token_header))
    return True



import json

from django.http import HttpResponseNotFound, HttpResponseNotAllowed, HttpResponse
from django.utils import timezone
from django.utils.decorators import method_decorator
from django.views import View
from django.views.decorators.csrf import csrf_exempt

from listener.models import Source
from listener.utils import authenticate


# Create your views here.


class WebhookView(View):

    @method_decorator(csrf_exempt)
    def dispatch(self, request, *args, **kwargs):
        slug = kwargs.get("slug")
        if slug:
            source = Source.objects.filter(slug=slug).first()
            if source:
                return self.process_webhook(request, source, *args, **kwargs)
        return HttpResponseNotFound("Invalid webhook slug")

    def process_webhook(self, request, source, *args, **kwargs):
        method = request.method.lower()
        if method != source.http_method.lower():
            return HttpResponseNotAllowed([source.http_method.lower()], "Invalid HTTP method")

        if not authenticate(request, source):
            return HttpResponse("Unauthorized", status=401)

        source.last_checked = timezone.now()
        source.save()

        body = request.body
        if hasattr(body, "decode"):
            body = body.decode("utf-8")
        try:
            data = json.loads(body)
            unique_id_field = source.unique_id_field
            if unique_id_field:
                unique_id = data
                for field in unique_id_field.split('.'):
                    unique_id = unique_id.get(field)
            else:
                unique_id = data.get('id')
            if unique_id:
                event = source.events.filter(unique_id=unique_id).first()
                if not event:
                    event = source.events.create(
                        name=source.name,
                        body=body,
                        headers=json.dumps(dict(request.headers)),
                        query_params=json.dumps(dict(request.GET)),
                        unique_id=unique_id,
                    )
                    event.save()
                if not event.processed:
                    processor_class = source.get_processor_class()
                    processor = processor_class(event)
                    processor.process()
            return HttpResponse("OK")
        except json.JSONDecodeError:
            return HttpResponse("Invalid JSON payload", status=400)



from setuptools import setup, find_packages


def read(f):
    return open(f, 'r', encoding='utf-8').read()


setup(
    name="django-listener",
    version='0.1',
    description="Automatically listens and processes webhooks from third party services",
    long_description=read("README.md"),
    long_description_content_type='text/markdown',
    url="https://github.com/sirrobot01/django-listener",
    author="Mukhtar Akere",
    author_email="akeremukhtar10@gmail.com",
    license="BSD-3-Clause",
    classifiers=[
        "Environment :: Web Environment",
        "Framework :: Django",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
    ],
    install_requires=[
        "django>=2.2"
    ],
    packages=find_packages(),
    include_package_data=True,
    extras_require={},
    project_urls={
        'Source': 'https://github.com/sirrobot01/django-listener',
    },
)
